"""this contains the template-specific entities' classes"""

from cubicweb_expense.entities import ExpenseLine as BaseExpenseLine


class ExpenseLine(BaseExpenseLine):
    @property
    def workcase(self):
        rql = "Any R WHERE E has_lines EL, EL eid %(el)s, E spent_for W, W ref R"
        rset = self._cw.execute(rql, {"el": self.eid})
        if rset:
            return rset[0][0]
        return None



"""accounting views for fresh template

:organization: Logilab
:copyright: 2008-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""

__docformat__ = "restructuredtext en"


from logilab.mtconverter import xml_escape

from cubicweb import _
from cubicweb.predicates import is_instance
from cubicweb_web.view import EntityView


class ExpenseAccountingXmlView(EntityView):
    __regid__ = "accexpense"
    __select__ = is_instance("Expense")

    title = _("accounting entry view")
    templatable = False
    content_type = "text/xml"

    def call(self):
        """display a list of entities by calling their <item_vid> view"""
        self.w('<?xml version="1.0" encoding="%s"?>\n' % self._cw.encoding)
        self.w(
            '<?xml-stylesheet href="%saccounting-entries.xsl" '
            'rel="stylesheet" type="text/xsl"?>\n' % self._cw.datadir_url
        )
        self.w("<ecritures>\n")
        for i in range(self.cw_rset.rowcount):
            self.cell_call(i, 0)
        self.w("</ecritures>\n")

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        rset = entity.related("has_lines")
        for i in range(len(rset)):
            self.wview("accentry", rset, row=i, col=0)


class ExpenseLineAccountingEntryXmlView(EntityView):
    __regid__ = "accentry"
    __select__ = is_instance(
        "ExpenseLine",
    )

    title = _("accounting entry view")
    templatable = False
    content_type = "text/xml"

    def call(self):
        """display a list of entities by calling their <item_vid> view"""
        self.w('<?xml version="1.0" encoding="%s"?>\n' % self._cw.encoding)
        self.w(
            '<?xml-stylesheet href="%saccounting-entries.xsl" '
            'rel="stylesheet" type="text/xsl"?>\n' % self._cw.datadir_url
        )
        self.w("<ecritures>\n")
        for i in range(self.cw_rset.rowcount):
            self.cell_call(i, 0)
        self.w("</ecritures>\n")

    def cell_call(self, row, col):
        entity = self.cw_rset.complete_entity(row, col)
        self.w('  <ecriture date="%s">\n' % entity.diem.strftime("%Y-%m-%d"))
        self.w("    <libelle>%s</libelle>\n" % xml_escape(entity.dc_long_title()))
        amount = round(entity.euro_amount(), 2)
        taxes = round(entity.taxes, 2)
        account = (
            entity.paid_by[0].account and xml_escape(entity.paid_by[0].account) or ""
        )
        self.w('    <credit compte="%s" montant="%.2f" />\n' % (account, amount))
        if entity.taxes:
            # XXX hardcoded account for VAT
            self.w('    <debit compte="44566" montant="%.2f" />\n' % entity.taxes)
        taxfree = int(round((amount - taxes) * 100))
        accounts = list(entity.paid_for)
        debit_quotient = taxfree / len(accounts)
        debit_remainder = taxfree % len(accounts)
        for account in accounts:
            if debit_remainder > 0:
                debit = (debit_quotient + 1) / 100.0
                debit_remainder -= 1
            else:
                debit = debit_quotient / 100.0
            account = account.account and xml_escape(account.account) or ""
            self.w('    <debit compte="%s" montant="%.2f" />\n' % (account, debit))
        if entity.workcase:
            self.w("    <groupe>%s</groupe>\n" % xml_escape(entity.workcase))
        self.w("  </ecriture>\n")



if "Training" not in schema:
    add_entity_type("Training")

# workflows doen't understand yams inheritance
rql(
    "SET WF workflow_of ET, ET default_workflow WF "
    'WHERE WF workflow_of WC, WC name "Workcase", ET name "Training"'
)
commit()



from os import makedirs
from os.path import join, exists

from cubicweb.server import hook
from cubicweb.server.sources import storages


class ServerStartupHook(hook.Hook):
    __regid__ = "drh.serverstartup"
    events = ("server_startup", "server_maintenance")

    def __call__(self):
        bfssdir = join(self.repo.config.appdatahome, "bfss")
        if not exists(bfssdir):
            makedirs(bfssdir)
            print("created", bfssdir)
        storage = storages.BytesFileSystemStorage(bfssdir)
        storages.set_attribute_storage(self.repo, "File", "data", storage)



"""specific actions for fresh template

:organization: Logilab
:copyright: 2008-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""

__docformat__ = "restructuredtext en"

from cubicweb import _
from cubicweb.predicates import is_instance
from cubicweb_web.action import Action


class AccountingAction(Action):
    __regid__ = "accaction"
    __select__ = is_instance("Expense")
    title = _("generate accounting entries")

    def url(self):
        entity = self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0)
        return entity.absolute_url(vid="accexpense")



"""template-specific forms/views/actions/components"""

from cubicweb_web import formwidgets as fw
from cubicweb_web.views import uicfg
from cubicweb_web.views.ajaxcontroller import ajaxfunc

_afs = uicfg.autoform_section
_affk = uicfg.autoform_field_kwargs
_afs.tag_subject_of(("Expense", "spent_for", "*"), "main", "attributes")
_afs.tag_subject_of(("Expense", "spent_for", "*"), "muledit", "attributes")
_affk.tag_subject_of(
    ("Expense", "spent_for", "*"),
    {
        "widget": fw.LazyRestrictedAutoCompletionWidget(
            autocomplete_initfunc="get_concerned_by",
            autocomplete_settings={"limit": 100, "delay": 300},
        )
    },
)


@ajaxfunc(output_type="json")
def get_concerned_by(self):
    term = self._cw.form["q"]
    limit = self._cw.form.get("limit", 50)
    return [
        {"value": eid, "label": ref}
        for eid, ref in self._cw.execute(
            "DISTINCT Any W,R ORDERBY R LIMIT %s WHERE W ref R,"
            "W ref ILIKE %%(term)s" % limit,
            {"term": "%%%s%%" % term},
        )
    ]



from six import text_type
import cwclientlib
from cubicweb.server.sources import datafeed
from cubicweb.dataimport import stores, importer


# XXX class copied in crm/fresh, should be kepts sync
class CWClientLibDataFeedParser(datafeed.DataFeedParser):
    """Base class for parsers that search for distant entities using cwclientlib."""

    def process_urls(self, *args, **kwargs):
        """IDataFeedParser main entry point."""
        self._source_uris = {}
        error = super(CWClientLibDataFeedParser, self).process_urls(*args, **kwargs)
        if not error:
            self.handle_deletion()
        return error

    def process(self, url, raise_on_error=False):
        """Called once by process_urls (several URL are not expected with this parser)."""
        # ensure url ends with a single slash for proper extid generation
        url = url.rstrip("/") + "/"
        eeimporter = self.build_importer(raise_on_error)
        entities = self.extentities_generator(url)
        set_cwuri = importer.use_extid_as_cwuri(eeimporter.extid2eid)
        eeimporter.import_entities(set_cwuri(entities))
        self.stats["created"] = eeimporter.created
        self.stats["updated"] = eeimporter.updated

    def build_importer(self, raise_on_error):
        """Instantiate and configure an importer"""
        etypes, extid2eid = self.init_extid2eid()
        existing_relations = self.init_existing_relations()
        store = stores.NoHookRQLObjectStore(
            self._cw, metagen=stores.MetadataGenerator(self._cw, source=self.source)
        )
        return importer.ExtEntitiesImporter(
            self._cw.vreg.schema,
            store,
            extid2eid=extid2eid,
            existing_relations=existing_relations,
            etypes_order_hint=etypes,
            import_log=self.import_log,
            raise_on_error=raise_on_error,
        )

    def handle_deletion(self, *args, **kwargs):
        for extid, (eid, etype) in self._source_uris.items():
            self._cw.entity_from_eid(eid, etype).cw_delete()

    def existing_entities(self, etype):
        rset = self._cw.execute(
            "Any XURI, X WHERE X cwuri XURI, X is {0},"
            " X cw_source S, S eid %(s)s".format(etype),
            {"s": self.source.eid},
        )
        for extid, eid in rset:
            self._source_uris[extid] = (eid, etype)
            yield extid, eid

    def states_map(self, etype):
        return dict(
            self._cw.execute(
                "Any SN,S WHERE S name SN, S state_of WF, "
                "ET default_workflow WF, ET name %(etype)s",
                {"etype": etype},
            )
        )

    def ext_entity(self, url, etype, eid, values):
        extid = url + text_type(eid)
        self._source_uris.pop(extid, None)
        return importer.ExtEntity(etype, extid, values)


class DataFeedFreshActivity(CWClientLibDataFeedParser):
    """Parser to import workcases from crm."""

    __regid__ = "fresh.workcases"

    def init_extid2eid(self):
        # put state eids in extid2eid as we'll want to link to them
        extid2eid = dict(self._cw.execute("Any X,X WHERE X is State"))
        # map existing orders and workorders from our souce
        etypes = ("Workcase",)
        for etype in etypes:
            for extid, eid in self.existing_entities(etype):
                extid2eid[extid] = eid
        return etypes, extid2eid

    def init_existing_relations(self):
        existing_relations = {}
        rset = self._cw.execute(
            "Any O,OS WHERE O in_state OS, O is Workcase, O cw_source S, S eid %(s)s",
            {"s": self.source.eid},
        )
        existing_relations["in_state"] = set(tuple(x) for x in rset)
        return existing_relations

    def extentities_generator(self, url):
        # connect to the instance using cwclientlib
        proxy = cwclientlib.cwproxy_for(url)
        # XXX check modification_date > last_update
        # information necessary to relate order to a state
        states_map = self.states_map("Workcase")

        for args in proxy.execute(
            "Any W,SN,WR,WS WHERE W is_instance_of Workcase, "
            "W ref WR, W subject WS, W in_state S, S name SN"
        ):
            eid = args.pop(0)
            state = args.pop(0)
            values = values_dict(args, ["ref", "subject"])
            values["in_state"] = set([states_map[state]])
            yield self.ext_entity(url, "Workcase", eid, values)


def values_dict(values_list, attributes):
    values = {}
    for k, v in zip(attributes, values_list):
        if v is not None:
            values[k] = set([v])
    return values



# -*- coding: utf-8 -*-
# postcreate script. You could setup a workflow here for example

for login in (
    "alf",
    "syt",
    "nico",
    "jphc",
    "ocy",
    "auc",
    "katia",
    "graz",
    "dede",
    "juj",
    "ludal",
    "steph",
    "arthur",
    "david",
    "joel",
    "gaston",
    "adim",
):
    rql(
        "INSERT CWUser E: E login %(login)s, E upassword %(login)s, E in_group G "
        'WHERE G name "users"',
        {"login": login},
    )
    rql(
        "INSERT PaidByAccount P: P label %(label)s, P associated_to U WHERE U login %(login)s",
        {"label": "refund account - %s" % login, "login": login},
    )
    rql(
        "INSERT PaidForAccount P: P label %(label)s",
        {"label": "charge account - %s" % login},
    )


for label in (
    "Logilab - CB Nicolas",
    "Logilab - CB Alexandre",
    "Logilab - CB Olivier",
    "Logilab - Espèces",
):
    rql("INSERT PaidByAccount P: P label %(label)s", {"label": label})



"""template automatic tests"""

import unittest

from cubicweb.devtools.htmlparser import XMLValidator
from cubicweb_web.devtools.testlib import AutomaticWebTest


class AutomaticWebTest(AutomaticWebTest):
    vid_validators = AutomaticWebTest.vid_validators.copy()
    vid_validators.update(
        {
            "accexpense": XMLValidator,
            "accentry": XMLValidator,
        }
    )


if __name__ == "__main__":
    unittest.main()



# pylint: disable-msg=W0622
"""cubicweb-fresh application packaging information"""

modname = "fresh"
distname = "cubicweb-fresh"

numversion = (1, 5, 3)
version = ".".join(str(num) for num in numversion)

license = "LGPL"
author = "Logilab"
author_email = "contact@logilab.fr"
description = "expense tracking application built on the CubicWeb framework"
web = "http://www.cubicweb.org/project/%s" % distname
classifiers = [
    "Environment :: Web Environment",
    "Framework :: CubicWeb",
    "Programming Language :: Python",
    "Programming Language :: JavaScript",
]

__depends__ = {
    "cubicweb[s3]": ">=4.8.0,<5.0.0",
    "cubicweb-api": ">=0.12.0,<0.13.0",
    "cubicweb-expense": ">=1.0.0,<2.0.0",
    "cubicweb-workcase": ">=1.0.0,<2.0.0",
    "cubicweb-searchui": ">=1.0.0,<2.0.0",
    "cubicweb-prometheus": ">=0.5.0,<0.6.0",
    "cubicweb-rqlcontroller": ">=1.0.0,<2.0.0",
    "cubicweb-signedrequest": ">=3.0.0,<4.0.0",
    "cubicweb-sentry": ">=1.0.0,<2.0.0",
    "cubicweb-card": ">=2.0.0,<3.0.0",
    "cubicweb-oauth2": ">=1.0.0,<2.0.0",
    "cubicweb-web": ">=1.0.0,<2.0.0",
    "reportlab": ">=3.6.0,<3.7.0",
    "cwclientlib": ">=1.3.0,<1.4.0",
}



# template's specific schema
from yams.buildobjs import RelationDefinition
from cubicweb.schema import RRQLExpression
from cubicweb_workcase.schema import Workcase


class Training(Workcase):
    __specializes_schema__ = True


class spent_for(RelationDefinition):
    subject = "Expense"
    object = "Workcase"
    cardinality = "?*"
    __permissions__ = {
        "read": ("managers", "users"),
        "add": ("managers", RRQLExpression('NOT (S in_state ST, ST name "accepted")')),
        "delete": (
            "managers",
            RRQLExpression('NOT (S in_state ST, ST name "accepted")'),
        ),
    }
